@@ -18,3 +18,18 @@ def register_models(register):
 @hookspec
 def register_embedding_models(register):
     "Register additional model instances that can be used for embedding"
+
+
+@hookspec
+def register_template_loaders(register):
+    "Register additional template loaders with prefixes"
+
+
+@hookspec
+def register_fragment_loaders(register):
+    "Register additional fragment loaders with prefixes"
+
+
+@hookspec
+def register_tools(register):
+    "Register functions that can be used as tools by the LLMs"
